from datetime import datetime
from typing import Optional
from uuid import UUID
from pydantic import BaseModel
from talentro.vacancies.dataclasses import VacancyInfo

from ..acquisition.models import ChannelType, CampaignGoal, Campaign as CampaignModel, Ad as AdModel, AdStatus
from ..general.dataclasses import ResolvableCompanyModel
from ..integrations.dataclasses import LinkInfo
from ..services.clients import MSClient
from ..vacancies.dataclasses import FeedInfo


# Campaign object
class CampaignInfo(ResolvableCompanyModel):
    id: UUID
    created_at: datetime
    updated_at: Optional[datetime]
    organization: UUID

    name: str
    external_id: Optional[str]
    status: str
    last_sync_date: Optional[datetime]
    ad_count: int
    auto_sync: bool
    channel: LinkInfo
    channel_type: ChannelType
    campaign_goal: Optional[CampaignGoal]
    feed: FeedInfo
    selection_criteria: Optional[dict]
    application_flow_settings: Optional[dict]
    status_detail: Optional[str]

    @staticmethod
    async def resolve_object(object_id: UUID, organization_id: UUID) -> "CampaignInfo | None":
        result = await MSClient.acquisition().get(f"campaigns/{object_id}/resolve", headers={"X-Organization-ID": str(organization_id)})

        if result.status_code != 200:
            return None

        return CampaignInfo(**result.json())

    @classmethod
    async def from_model(cls: "CampaignInfo", model: CampaignModel) -> 'CampaignInfo':
        channel = await LinkInfo.resolve_object(model.channel_id, model.organization)
        feed = await FeedInfo.resolve_object(model.feed_id, model.organization)

        return cls(
            id=model.id,
            created_at=model.created_at,
            updated_at=model.updated_at,
            organization=model.organization,
            name=model.name,
            external_id=model.external_id,
            status=model.status,
            last_sync_date=model.last_sync_date,
            ad_count=model.ad_count,
            auto_sync=model.auto_sync,
            channel=channel,
            channel_type=model.channel_type,
            campaign_goal=model.campaign_goal,
            feed=feed,
            selection_criteria=model.selection_criteria,
            application_flow_settings=model.application_flow_settings,
            status_detail=model.status_detail
        )


class CampaignConfig(BaseModel):
    id: UUID
    created_at: datetime
    updated_at: Optional[datetime]
    organization: UUID

    name: str
    external_id: Optional[str]
    status: str
    last_sync_date: Optional[datetime]
    ad_count: int
    auto_sync: bool
    channel_id: UUID
    channel_type: ChannelType
    campaign_goal: Optional[CampaignGoal]
    feed_id: UUID
    selection_criteria: Optional[dict]
    application_flow_settings: Optional[dict]

    @classmethod
    async def from_model(cls: "CampaignConfig", model: CampaignModel) -> 'CampaignConfig':
        return cls(**model.model_dump())


class AdInfo(ResolvableCompanyModel):
    id: UUID
    created_at: datetime
    updated_at: Optional[datetime]
    organization: UUID

    name: str
    external_id: Optional[str]
    status: AdStatus
    primary_text: str
    title: str
    description: Optional[str]
    status_detail: Optional[str]
    conversion_goal: Optional[str]
    vacancy_id: UUID

    campaign: CampaignInfo

    @staticmethod
    async def resolve_object(object_id: UUID, organization_id: str) -> "ResolvableCompanyModel":
        pass

    @classmethod
    async def from_model(cls: "AdInfo", model: AdModel) -> 'AdInfo':
        campaign = await CampaignInfo.from_model(model.campaign)

        return cls(
            id=model.id,
            created_at=model.created_at,
            updated_at=model.updated_at,
            organization=model.organization,

            name=model.name,
            external_id=model.external_id,
            status=model.status,
            primary_text=model.primary_text,
            title=model.title,
            description=model.description,
            status_detail=model.status_detail,
            conversion_goal=model.conversion_goal,
            vacancy_id=model.vacancy_id,

            campaign=campaign,
        )


from threading import Lock


class SingletonMeta(type):
    _instances = {}
    _lock: Lock = Lock()

    def __call__(cls, *args, **kwargs):
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]


class Singleton:
    _instances = {}

    @staticmethod
    def singleton_instance(class_obj, *args, **kwargs):
        if class_obj not in Singleton._instances:
            Singleton._instances[class_obj] = class_obj(*args, **kwargs)
        return Singleton._instances[class_obj]



import asyncio
import os
from typing import List, Tuple, Optional, Callable

import aio_pika
from aiormq import AMQPConnectionError

from ..event import Message, Event
from ..util.singleton import SingletonMeta


class RabbitMQ:
    def __init__(self):
        self.connection = None
        self.channel = None
        self._on_reconnect_callbacks = []

    async def connect(self):
        self.connection = await aio_pika.connect_robust(
            host=os.getenv('QUEUE_HOST'),
            port=int(os.getenv('QUEUE_PORT')),
            login=os.getenv('QUEUE_USER'),
            password=os.getenv('QUEUE_PASS'),
            virtualhost=os.getenv("QUEUE_VHOST", "/"),
        )
        print("  RabbitMQ - Connected!")
        self.channel = await self.connection.channel()
        await self.channel.set_qos(prefetch_count=int(os.getenv("QUEUE_PREFETCH", "16")))

        # Registreer de reconnect callback op het kanaal, niet de connectie
        self.channel.reopen_callbacks.add(self._on_channel_reopen)
        print("  RabbitMQ - Channel created & QoS set!")

    async def _on_channel_reopen(self, channel):
        print("  RabbitMQ - Channel restored, re-applying settings...")
        await channel.set_qos(prefetch_count=int(os.getenv("QUEUE_PREFETCH", "16")))
        for callback in self._on_reconnect_callbacks:
            await callback()

    async def setup_topology(
        self,
        exchange_name: str,
        exchange_type: aio_pika.ExchangeType = aio_pika.ExchangeType.TOPIC,
        durable: bool = True,
        bindings: List[Tuple[str, str]] = (),
        dlx_name: Optional[str] = None
    ):
        exchange = await self.channel.declare_exchange(
            exchange_name, exchange_type, durable=durable
        )
        print(f"  RabbitMQ - Exchange declared: {exchange_name}")

        # Create queues and bindings
        for queue_name, routing_key in bindings:
            args = {}
            if dlx_name:
                args["x-dead-letter-exchange"] = dlx_name
                args["x-dead-letter-routing-key"] = f"{queue_name}.dead"

            queue = await self.channel.declare_queue(
                queue_name, durable=True, arguments=args
            )
            await queue.bind(exchange, routing_key=routing_key)
            print(f"  RabbitMQ - Queue bound: {queue_name} ←({routing_key})– {exchange_name}")

        # Create DLX
        if dlx_name:
            await self.channel.declare_exchange(dlx_name, aio_pika.ExchangeType.DIRECT, durable=True)

            # Dead-letter queues
            for queue_name, _ in bindings:
                dlq = await self.channel.declare_queue(f"{queue_name}.dlq", durable=True)
                await dlq.bind(dlx_name, routing_key=f"{queue_name}.dead")
                print(f"  RabbitMQ - DLQ bound: {dlq.name} ← {dlx_name}")

    async def consume(self, exchange_name: str, queue: str, callback: Callable, dlx_name: Optional[str] = None):
        while True:
            try:
                print(f"  RabbitMQ - Consuming from queue: {queue}")

                args = {}
                if dlx_name:
                    args["x-dead-letter-exchange"] = dlx_name
                    args["x-dead-letter-routing-key"] = f"{queue}.dead"

                queue_object = await self.channel.declare_queue(queue, durable=True, arguments=args)

                # Bind to exchange if not already bound
                if exchange_name != "default":
                    exchange = await self.channel.get_exchange(exchange_name)
                    await queue_object.bind(exchange, routing_key=queue)

                async for message in queue_object:
                    async with message.process():
                        event: Event = Event.decode(message.body)
                        print(f"  RabbitMQ ({queue}) - [x] Received event: {event.meta.event_type} (trace: {event.meta.trace_id})")
                        await callback(queue, event)
            except (aio_pika.exceptions.AMQPConnectionError, aio_pika.exceptions.ChannelClosed):
                print(f"  RabbitMQ - Connection lost while consuming {queue}. Retrying in 5 seconds...")
                await asyncio.sleep(5)
            except Exception as e:
                print(f"  RabbitMQ - Unexpected error in consumer {queue}: {e}")
                await asyncio.sleep(5)

    async def send_message(self, message: Message):
        # Haal de exchange op
        if message.exchange == "default":
            exchange = self.channel.default_exchange
            routing_key = message.routing_key
        else:
            exchange = await self.channel.get_exchange(message.exchange)
            routing_key = message.routing_key

        # Bouw aio_pika.Message met alle metadata
        amqp_message = aio_pika.Message(
            body=message.body(),
            headers=message.merged_headers(),
            delivery_mode=aio_pika.DeliveryMode.PERSISTENT if message.persistent else aio_pika.DeliveryMode.NOT_PERSISTENT,
        )

        await exchange.publish(amqp_message, routing_key=routing_key)
        print(
            f"  RabbitMQ - [x] Sent event: {message.event.meta.event_type} "
            f"to exchange '{message.exchange}' with routing key '{routing_key}' "
            f"(trace: {message.event.meta.trace_id})"
        )


class QueueContext(metaclass=SingletonMeta):

    def __init__(self):
        self._rabbit_mq = RabbitMQ()
        self._message_callbacks = []

    async def connect(self):
        try:
            await self._rabbit_mq.connect()
        except AMQPConnectionError as e:
            print(f"  RabbitMQ - Connection failed: {e}")
            await asyncio.sleep(5)
            await self.connect()


    async def setup_topology(
        self,
        exchange_name: str,
        exchange_type: aio_pika.ExchangeType = aio_pika.ExchangeType.TOPIC,
        durable: bool = True,
        bindings: List[Tuple[str, str]] = (),
        dlx_name: Optional[str] = None
    ):
        await self._rabbit_mq.setup_topology(
            exchange_name, exchange_type, durable, bindings, dlx_name
        )

    async def start_consuming(self, exchange_name: str, queue: str, dlx_name: Optional[str] = None):
        asyncio.create_task(
            self._rabbit_mq.consume(exchange_name, queue, self._on_new_message, dlx_name=dlx_name)
        )

    async def _on_new_message(self, queue: str, event: Event):
        for callback_object in self._message_callbacks:
            if callback_object.get("queue") == queue:
                await callback_object.get("callback")(event)

    def add_on_message_callback(self, queue: str, callback: Callable):
        self._message_callbacks.append({'queue': queue, 'callback': callback})

    async def send_message(self, message: Message):
        await self._rabbit_mq.send_message(message)